// Projector.cpp
//

// LDraw Projector
// Map a flat LDraw file (pattern) onto a raised one.
// The mapping is done by projection along Y axis 
// (the pattern and the raised shape must both be mostly in X-Z plane)
// Projected items are lines, triangle and quads.
// Limited support of flat primitives (discs, rings...)
// Primitives used in raised file must be inlined prior to use
// For good results, each projected element as well as its origin must
// reside entirely onto a single triangle/quad of raised shape.

// Usage: projector LdrawPatternFile Ldraw3DFile Ldraw3DPatternFileOut

// Philippe "Philo" Hurbain - 2008 - www.philohome.com

// Contains excerpts of triangle intersection code from Tomas Mller
// See thesis here: http://www.cs.lth.se/home/Tomas_Akenine_Moller/pubs/tritri.pdf
// Complete code here: http://jgt.akpeters.com/papers/Moller97/tritri.html


// Contains softSurfer Line - Triangle intersection (adaptation)
// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

#include "stdafx.h"

/* Triangle/triangle intersection test routine,
 * by Tomas Moller, 1997.
 * See article "A Fast Triangle-Triangle Intersection Test",
 * Journal of Graphics Tools, 2(2), 1997
 * updated: 2001-06-20 (added line of intersection)
 *
 * This version computes the line of intersection as well (if they are not coplanar):
 * int tri_tri_intersect_with_isectline(double V0[3],double V1[3],double V2[3], 
 *				        double U0[3],double U1[3],double U2[3],int *coplanar,
 *				        double isectpt1[3],double isectpt2[3]);
 * coplanar returns whether the tris are coplanar
 * isectpt1, isectpt2 are the endpoints of the line of intersection
 */

#include <math.h>


#define USE_EPSILON_TEST TRUE  
#define EPSILON 0.001

/* Null vector */

double null[3] = {0,0,0};

/* some macros */

#define CROSS(dest,v1,v2)                      \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];

#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

#define SUB(dest,v1,v2) dest[0]=v1[0]-v2[0]; dest[1]=v1[1]-v2[1]; dest[2]=v1[2]-v2[2]; 

#define ADD(dest,v1,v2) dest[0]=v1[0]+v2[0]; dest[1]=v1[1]+v2[1]; dest[2]=v1[2]+v2[2]; 

#define MULT(dest,v,factor) dest[0]=factor*v[0]; dest[1]=factor*v[1]; dest[2]=factor*v[2];

#define SET(dest,src) dest[0]=src[0]; dest[1]=src[1]; dest[2]=src[2]; 

#define MANHATTAN(v1, v2) (fabs(v1[0]-v2[0]) + fabs(v1[1]-v2[1]) + fabs(v1[2]-v2[2]))


int tri_segment_intersect_with_isectpoint(double V0[3],double V1[3],double V2[3],
										  double S0[3],double S1[3],
										  double isectpt[3])
{
  double u[3],v[3],n[3]; // triangle vectors
  double dir[3], w0[3], w[3];// ray vectors
  double temp[3];
  double r, a, b;
  double d1;
  double du0,du1;
  double du0du1;
  
  /* compute plane equation of triangle(V0,V1,V2) */
  SUB(u,V1,V0);
  SUB(v,V2,V0);
  CROSS(n,u,v);
  d1=-DOT(n,V0);
  /* plane equation 1: n.X+d1=0 */

  /* put S0, S1 into plane equation 1 to compute signed distances to the plane*/
  du0=DOT(n,S0)+d1;
  du1=DOT(n,S1)+d1;
  
  /* coplanarity robustness check */
#if USE_EPSILON_TEST==TRUE
  if(fabs(du0)<EPSILON) du0=0.0;
  if(fabs(du1)<EPSILON) du1=0.0;
#endif
  du0du1=du0*du1;

//  if(du0du1>0.0f) /* same sign on both of them + not equal 0 ? */
//    return 0;                    /* no intersection occurs */

  if(MANHATTAN(n,null)<EPSILON) // triangle is degenerate
	return 0;

//  *idx=(du0>0)?0:1; // Which segment end is "outside" triangle?

  SUB(dir, S1, S0);  // ray direction vector
  SUB(w0, S0, V0);
  a=-DOT(n,w0);
  b=DOT(n,dir);

  if (fabs(b) < EPSILON) {     // ray is parallel to triangle plane
     return 0;             
  }

  r = a / b;
  MULT(temp, dir, r);
  ADD(isectpt, S0, temp);

  // is I inside T?
    double uu, uv, vv, wu, wv, D1;
    uu = DOT(u,u);
    uv = DOT(u,v);
    vv = DOT(v,v);
    SUB(w, isectpt, V0); 
    wu = DOT(w,u);
    wv = DOT(w,v);
    D1 = uv * uv - uu * vv;

    // get and test parametric coords
    double s, t;
    s = (uv * wv - vv * wu) / D1;
    if (s < -EPSILON || s > 1.0+EPSILON)        // I is outside T
        return 0;
    t = (uv * wu - uu * wv) / D1;
    if (t < -EPSILON || (s + t) > 1.0+EPSILON)  // I is outside T
        return 0;

    return 1;                      // I is in T
}

#define MAX_TRI 1000
double InTri[MAX_TRI+1][3][3];
int NumTri;

void project(double u[3], double proj[3], int *idx)
{
	double u1[3], u2[3];
	double isec[3];
	int FlagInter;
	int i;

	SET(u1, u);
	SET(u2, u);
	u1[1]=0;
	u2[1]=1;
	for(i=0; i<NumTri; i++)
	{
		FlagInter = tri_segment_intersect_with_isectpoint(InTri[i][0], InTri[i][1], InTri[i][2],
										  u1,u2,
										  isec);
		if(FlagInter == 1) break;
	}
	if (FlagInter == 1)
	{
		SET (proj, isec);
		*idx = i;
	}
	else 
	{
		SET (proj, u1);
		*idx=0;
	}
}


int main(int argc, char* argv[])
{

	double InForm[4][3];

	FILE *LdrawPatternFile;
	FILE *Ldraw3DFile;
	FILE *Ldraw3DPatternFileOut;

	int NumLine;
	double u[3], v[3], n[3], m[3];

	char buf[1024];
	char prim[100];
	int type, color;

	int i;
	
	int index;

	int flag=0;

	double proj[3];

	printf("\nLdraw Projector v1.2 - by Philo");
	printf("\n-------------------------------\n\n");

	if(argc != 4) {
		printf ("Usage: projector LdrawPatternFile Ldraw3DFile Ldraw3DPatternFileOut\n");
		printf("Press <Enter> to quit");
		getchar();
		return 1;
	}
	

	LdrawPatternFile = fopen (argv[1], "rt");
	if(LdrawPatternFile == NULL)
	{
		printf("Can't open file\n%s\n", argv[1]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	Ldraw3DFile = fopen (argv[2], "rt");
	if(Ldraw3DFile == NULL)
	{
		printf("Can't open file\n%s\n", argv[2]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	Ldraw3DPatternFileOut = fopen (argv[3], "wt");
	if(Ldraw3DPatternFileOut == NULL)
	{
		printf("Can't create file\n%s\n", argv[3]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}


	printf("\nRead 3D shape file...\n");
	NumTri=0;
	while (fgets(buf, 1024, Ldraw3DFile))
	{
		type = -1;
		sscanf(buf, "%d", &type);
		if(type==3)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InTri[NumTri][0][0], &InTri[NumTri][0][1], &InTri[NumTri][0][2], 
				&InTri[NumTri][1][0], &InTri[NumTri][1][1], &InTri[NumTri][1][2],
				&InTri[NumTri][2][0], &InTri[NumTri][2][1], &InTri[NumTri][2][2]);
			NumTri++;
//			fprintf(LdrawTrimmedLineFileOut, "%s",buf);	// DEBUG
		}
		if(type==4)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InTri[NumTri][0][0], &InTri[NumTri][0][1], &InTri[NumTri][0][2], 
				&InTri[NumTri][1][0], &InTri[NumTri][1][1], &InTri[NumTri][1][2],
				&InTri[NumTri][2][0], &InTri[NumTri][2][1], &InTri[NumTri][2][2],
				&InTri[NumTri+1][2][0], &InTri[NumTri+1][2][1], &InTri[NumTri+1][2][2]);
//			fprintf(LdrawTrimmedLineFileOut, "3 24 %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",InTri[NumTri][0][0], InTri[NumTri][0][1], InTri[NumTri][0][2], InTri[NumTri][1][0], InTri[NumTri][1][1], InTri[NumTri][1][2], InTri[NumTri][2][0], InTri[NumTri][2][1], InTri[NumTri][2][2]);	// DEBUG


			SET(InTri[NumTri+1][0], InTri[NumTri][0]);
			SET(InTri[NumTri+1][1], InTri[NumTri][2]);

//			fprintf(LdrawTrimmedLineFileOut, "3 24 %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",InTri[NumTri+1][0][0], InTri[NumTri+1][0][1], InTri[NumTri+1][0][2], InTri[NumTri+1][1][0], InTri[NumTri+1][1][1], InTri[NumTri+1][1][2], InTri[NumTri+1][2][0], InTri[NumTri+1][2][1], InTri[NumTri+1][2][2]);	// DEBUG

			NumTri+=2;

		}

		if(NumTri>MAX_TRI)
		{
			printf("Too many triangles in file 2\n");
			printf("Press <Enter> to quit");
			getchar();
			return 3;
		}

	}
	printf("%d triangles in 3D shape file\n", NumTri);

	printf("\nRead and project pattern file...\n");
	NumLine=0;
	// fprintf(Ldraw3DPatternFileOut, "0 BFC CERTIFY CW\n");
	while (fgets(buf, 1024, LdrawPatternFile))
	{
		type = -1;
		sscanf(buf, "%d", &type);
		if(type==2)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InForm[0][0], &InForm[0][1], &InForm[0][2], 
				&InForm[1][0], &InForm[1][1], &InForm[1][2]);
			NumLine++;
		}
		else if (type==3)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InForm[0][0], &InForm[0][1], &InForm[0][2], 
				&InForm[1][0], &InForm[1][1], &InForm[1][2],
				&InForm[2][0], &InForm[2][1], &InForm[2][2]);
			NumLine++;
		}

		else if ((type==4))
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InForm[0][0], &InForm[0][1], &InForm[0][2], 
				&InForm[1][0], &InForm[1][1], &InForm[1][2],
				&InForm[2][0], &InForm[2][1], &InForm[2][2],
				&InForm[3][0], &InForm[3][1], &InForm[3][2]);
			NumLine++;
		}

		else if ((type==1))
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %s", &type, &color, 
				&InForm[0][0], &InForm[0][1], &InForm[0][2], 
				&InForm[1][0], &InForm[1][1], &InForm[1][2],
				&InForm[2][0], &InForm[2][1], &InForm[2][2],
				&InForm[3][0], &InForm[3][1], &InForm[3][2], &prim);
			NumLine++;
		}

		if((type>1) && (type<5))
		{
			fprintf(Ldraw3DPatternFileOut, "%d %d", type, color);
			for(i=0; i<type; i++)
			{
				project(InForm[i], proj, &index);
				fprintf(Ldraw3DPatternFileOut, " %.7lg %.7lg %.7lg", proj[0], proj[1], proj[2]);
			}
			fprintf(Ldraw3DPatternFileOut, "\n");
		}
		else if(type == 1)
		{
			double x,z;
			project(InForm[0], proj, &index);
			fprintf(Ldraw3DPatternFileOut, "%d %d %.7lg %.7lg %.7lg", type, color, proj[0], proj[1], proj[2]);
			// Calculate normal to surface at primitive center
			SUB(u, InTri[index][1], InTri[index][0]);
			SUB(v, InTri[index][1], InTri[index][2]);
			CROSS(m,u,v);
			MULT(n, m, (1/m[1]));

			// Y matrix coefficients
			x= -(InForm[1][0] * n[0] + InForm[3][0] * n[2]);
			z= -(InForm[1][2] * n[0] + InForm[3][2] * n[2]);

			//printf("\n%lf %lf %lf %lf %lf\n",n[0], n[1], n[2], x, z);
			fprintf(Ldraw3DPatternFileOut, " %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %s\n", InForm[1][0], InForm[1][1], InForm[1][2], x, n[1], z, InForm[3][0], InForm[3][1], InForm[3][2], prim);
		}

	}
	printf("%d lines in file 2\n", NumLine);

	fclose (LdrawPatternFile);
	fclose (Ldraw3DFile);
	fclose (Ldraw3DPatternFileOut);
	printf("Press <Enter> to quit");
	getchar();
	return(0);
}
